/*
 *  linux/fs/qnx4fs/bitmap.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

static inline void
clear_bits (u8 *ptr, int from, int len)
{
	for (; len; len--)
		ext2_clear_bit(from++, ptr);
}

static inline void
clear_low_bits (u8 *ptr, int len)
{
	*ptr &= (0xff << len);
}

static inline void
clear_high_bits (u8 *ptr, int len)
{
	*ptr &= (0xff >> len);
}

static inline void
clear_bytes (u8 *bitmap, int from, int to)
{
	memset(bitmap + from, 0, to - from);
}

static inline struct buffer_head *
bitmap_bread (kdev_t dev, unsigned long block)
{
	return bread(dev, block + QNX4FS_BITMAP_BLOCK - 1, QNX4FS_BLOCK_SIZE);
}

static inline void
clear_blocks (kdev_t dev, unsigned long from, unsigned long to)
{
	struct buffer_head *bp;

	for (; from < to; from++) {
		bp = bitmap_bread(dev, from);
		if (bp == NULL)
			continue;
		memset(bp->b_data, 0, QNX4FS_BLOCK_SIZE);
		mark_buffer_dirty(bp);
		brelse(bp);
	}
}

void
qnx4fs_bfree (struct super_block *sbp, unsigned long from, unsigned long len)
{
	kdev_t dev;
	struct buffer_head *bp;
	unsigned long to;
	unsigned int bfrom, bto;
	unsigned long block_from, block_to;
	u8 *bitmap;

	to = --from + len;
	if (len == 0 || to >= QNX4FS_SB(sbp)->s_blocks)
		return;

	block_from = from >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	block_to   = to >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	from      &= (QNX4FS_BLOCK_SIZE*8 - 1);
	to        &= (QNX4FS_BLOCK_SIZE*8 - 1);
	bfrom      = from >> 3;
	bto        = to >> 3;

	dev = sbp->s_dev;
	bp = bitmap_bread(dev, block_to);
	if (bp == NULL)
		return;
	bitmap = (u8 *)bp->b_data;
	switch (block_to - block_from) {
	case 0:
		switch (bto - bfrom) {
		case 0:
			clear_bits(&bitmap[bfrom], from & 7, len);
			break;
		default:
			clear_low_bits(&bitmap[bto], to & 7);
			clear_bytes(bitmap, (from + 7) >> 3, bto);
			clear_high_bits(&bitmap[bfrom], 8 - (from & 7));
			break;
		}
		break;
	default:
		clear_low_bits(&bitmap[bto], to & 7);
		clear_bytes(bitmap, 0, bto);
		mark_buffer_dirty(bp);
		brelse(bp);
		clear_blocks(dev, block_from + 1, block_to);
		bp = bitmap_bread(dev, block_from);
		if (bp == NULL)
			return;
		bitmap = (u8 *)bp->b_data;
		clear_bytes(bitmap, (from + 7) >> 3, QNX4FS_BLOCK_SIZE);
		clear_high_bits(&bitmap[bfrom], 8 - (from & 7));
		break;
	}
	mark_buffer_dirty(bp);
	brelse(bp);
}

int
qnx4fs_balloc (struct super_block *sbp, unsigned long *goal)
{
	struct buffer_head *bp;
	unsigned long block, nblocks;
	int nr, res;
	void *dp;

	block = (*goal - 1) >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	nr = (*goal - 1) & (QNX4FS_BLOCK_SIZE*8 - 1);
	bp = bitmap_bread(sbp->s_dev, block);
	dp = bp->b_data;
	if (bp) {
		if (test_bit (nr, dp)) {
			brelse (bp);
		} else {
			res = 1;
			goto found;
		}
	}
	block = nr = res = 0;
	nblocks = QNX4FS_SB(sbp)->s_bitmap_blocks;
	for (; block < nblocks; block++, nr = 0) {
		bp = bitmap_bread(sbp->s_dev, block);
		if (bp == NULL)
			continue;
		dp = bp->b_data;
		nr = find_next_zero_bit(dp, QNX4FS_BLOCK_SIZE*8, nr);
		if (nr < QNX4FS_BLOCK_SIZE*8) {
			*goal = (block<<(QNX4FS_BLOCK_SIZE_BITS + 3)) + 1 + nr;
			goto found;
		}
		brelse(bp);
	}
	return -ENOSPC;
found:
	__set_bit(nr, dp);
	mark_buffer_dirty(bp);
	brelse(bp);
	return res;
}

unsigned long
qnx4fs_count_free_blocks (struct super_block *sbp)
{
	struct buffer_head *bp;
	unsigned long count = 0;
	unsigned long nblocks, block;
	unsigned *w;
	int i;

	nblocks = QNX4FS_SB(sbp)->s_bitmap_blocks;
	for (block = 0; block < nblocks; block++) {
		bp = bitmap_bread (sbp->s_dev, block);
		if (bp == NULL)
			continue;
		w = (unsigned *)bp->b_data;
		for (i = QNX4FS_BLOCK_SIZE/4; i > 0; i--) {
			count += hweight32 (*w++);
		}
		brelse(bp);
	}
	return (QNX4FS_SB(sbp)->s_bitmap_blocks
		<< (QNX4FS_BLOCK_SIZE_BITS + 3)) - count;
}
