/*
 *  linux/fs/qnx4fs/common.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

inline void
qnx4fs_inc_nlink (struct inode *ip)
{
	ip->i_nlink++;
	mark_inode_dirty(ip);
}

inline void
qnx4fs_dec_nlink (struct inode *ip)
{
	ip->i_nlink--;
	mark_inode_dirty(ip);
}

inline unsigned long
ino2phys (ino_t ino)
{
	return ino >> QNX4FS_INODES_PER_BLOCK_BITS;
}

inline int
ino2slot (ino_t ino)
{
	return ino & (QNX4FS_INODES_PER_BLOCK-1);
}

inline ino_t
link2ino (qnx4fs_dirent_t *qdp)
{
	return direct2ino(le32_to_cpu(qdp->l.block), qdp->l.index);
}

inline ino_t
direct2ino (unsigned long phys, int slot)
{
	return ((phys-1) << QNX4FS_INODES_PER_BLOCK_BITS) + slot;
}

inline long
qnx4fs_nblocks (struct inode *ip)
{
	return (ip->i_size + QNX4FS_BLOCK_SIZE-1) >> QNX4FS_BLOCK_SIZE_BITS;
}

struct buffer_head *
qnx4fs_get_slot (struct inode *ip, ino_t *inop, qnx4fs_dirent_t **qdpp, int n)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long nblocks, block;
	int slot = 1;

	nblocks = qnx4fs_nblocks(ip);
	for (block = 0; block < nblocks; block++, slot = 0) {
		phys = qnx4fs_map_block(ip, block);
		if (!phys)
			continue;
		bp = bread(ip->i_dev, phys - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			continue;
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (qdp->i.status == 0) {
				*inop = direct2ino(phys, slot);
				*qdpp = qdp;
				goto found;
			}
		}
		brelse(bp);
	}
	bp = qnx4fs_iexpand_special(ip, n);
	if (bp == NULL)
		return ERR_PTR(-ENOSPC);
	*inop = direct2ino(bp->b_blocknr + 1, 0);
	*qdpp = (qnx4fs_dirent_t *)bp->b_data;
found:
	return bp;
}

void
qnx4fs_mark_slot (qnx4fs_dirent_t *qdp, struct buffer_head *bp, int status)
{
	qdp->i.status = status;
	if (!status)
		qdp->i.nlink = 0;
	mark_buffer_dirty(bp);
	brelse(bp);
}
