/*
 *  linux/fs/qnx4fs/ialloc.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/pagemap.h>
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

int
qnx4fs_dialloc (struct super_block *sbp, ino_t *inop)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;

	bp = qnx4fs_get_slot(QNX4FS_SB(sbp)->s_inodes, inop, &qdp,
			     QNX4FS_INODES_EXPAND_DELTA);
	if (IS_ERR(bp))
		return PTR_ERR(bp);
	qnx4fs_mark_slot(qdp, bp, QNX4FS_FILE_USED);
	return 0;
}

void
qnx4fs_difree (kdev_t dev, ino_t ino)
{
	struct buffer_head *bp;
	qnx4fs_dinode_t *qdp;

	bp = bread(dev, ino2phys(ino), QNX4FS_BLOCK_SIZE);
	if (bp == NULL) {
		BUG ();
		return;
	}
	qdp = (qnx4fs_dirent_t *)bp->b_data + ino2slot(ino);
	qnx4fs_mark_slot (qdp, bp, 0);
}

struct buffer_head *
qnx4fs_inew (struct inode *dip, struct dentry *dp,
	     struct inode **ipp, qnx4fs_dirent_t **qdpp)
{
	struct buffer_head *bp;
	struct inode *ip;
	ino_t ino;
	int err, longname = (dp->d_name.len > QNX4FS_SHORT_NAME_MAX);

	ip = new_inode(dip->i_sb);
	if (ip == NULL)
		return ERR_PTR(-ENOMEM);

	bp = qnx4fs_get_slot(dip, &ino, qdpp, QNX4FS_DIR_EXPAND_DELTA);
	if (IS_ERR(bp)) {
		qnx4fs_dec_nlink(ip);
		iput(ip);
		return bp;
	}
	if (longname) {
		err = qnx4fs_dialloc(dip->i_sb, &ino);
		if (err) {
			brelse(bp);
			qnx4fs_dec_nlink(ip);
			iput(ip);
			return ERR_PTR(err);
		}
	}

	*ipp			= ip;
	QNX4FS_I(ip)->i_linked	= longname;
	ip->i_ino		= ino;
	ip->i_uid		= current->fsuid;
	ip->i_gid		= (dip->i_mode & S_ISGID)
				  ? dip->i_gid : current->fsgid;
	ip->i_mtime		= ip->i_atime
				= ip->i_ctime
				= CURRENT_TIME;
	ip->i_blksize		= PAGE_CACHE_SIZE;
	ip->i_blocks		= 0;
	QNX4FS_I(ip)->i_xtnr	= -1;

	return bp;
}

static qnx4fs_xb_t *
qnx4fs_xalloc (struct inode *ip, struct buffer_head **bpp,
	       unsigned long *xbnrp)
{
	struct buffer_head *bp;
	qnx4fs_xb_t *xb;
	unsigned long goal = 1;

	if (qnx4fs_balloc(ip->i_sb, &goal))
		return NULL;
	bp = bread(ip->i_dev, goal - 1, QNX4FS_BLOCK_SIZE);
	if (bp == NULL) {
		qnx4fs_bfree(ip->i_sb, goal, 1);
		return NULL;
	}
	xb = (qnx4fs_xb_t *)bp->b_data;
	memset(xb, 0, QNX4FS_BLOCK_SIZE);
	xb->xb_prev	= 0;
	xb->xb_next	= 0;
	xb->xb_nxt	= 1;
	memcpy(xb->xb_signature, "IamXblk\0", 8);
	xb->xb_xt0.start = 0;
	xb->xb_xt0.size  = 0;
	++ip->i_blocks;

	*bpp = bp;
	*xbnrp = goal;

	return xb;
}

struct buffer_head *
qnx4fs_iexpand_special (struct inode *ip, int n)
{
	struct buffer_head *bp, *rbp;
	unsigned long phys;
	int i;

	phys = qnx4fs_iexpand(ip);
	if (!phys)
		return 0;
	rbp = bread(ip->i_dev, phys - 1, QNX4FS_BLOCK_SIZE);
	if (rbp == NULL)
		goto fail;
	memset(rbp->b_data, 0, QNX4FS_BLOCK_SIZE);
	for (i = n-1; i > 0; i--) {
		phys = qnx4fs_iexpand(ip);
		if (!phys)
			goto fail;
		bp = bread(ip->i_dev, phys - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			goto fail;
		memset(bp->b_data, 0, QNX4FS_BLOCK_SIZE);
		mark_buffer_dirty(bp);
		brelse(bp);
	}
	ip->i_size += n*QNX4FS_BLOCK_SIZE;
	mark_inode_dirty(ip);
	return rbp;
fail:
	qnx4fs_truncate(ip);
	return 0;
}

unsigned long
qnx4fs_iexpand (struct inode *ip)
{
	struct buffer_head *bp = NULL, *bp2 = NULL;
	qnx4fs_xb_t *xb = NULL;
	unsigned long goal, tmp, xbnr = QNX4FS_I(ip)->i_xbnr_last;
	int ixt = 0, res;

	switch (QNX4FS_I(ip)->i_nxt) {
	case 0:
		goal = 1;
		break;
	case 1:
		goal = QNX4FS_I(ip)->i_xt0.start + QNX4FS_I(ip)->i_xt0.size;
		break;
	default:
		bp = bread(ip->i_dev, xbnr - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			goto fail;
		xb = (qnx4fs_xb_t *)bp->b_data;
		ixt = xb->xb_nxt - 1;
		goal = le32_to_cpu(xb->xb_xts[ixt].start)
		     + le32_to_cpu(xb->xb_xts[ixt].size);
	}
	res = qnx4fs_balloc(ip->i_sb, &goal);
	if (res < 0)
		goto fail;
	switch (QNX4FS_I(ip)->i_nxt) {
	case 0:
		++QNX4FS_I(ip)->i_nxt;
		QNX4FS_I(ip)->i_xt0.start  = goal;
		QNX4FS_I(ip)->i_xt0.size   = 1;
		QNX4FS_I(ip)->i_phys       = goal;
		QNX4FS_I(ip)->i_offset_end = 1;
		break;
	case 1:
		if (res) {
			++QNX4FS_I(ip)->i_xt0.size;
			++QNX4FS_I(ip)->i_offset_end;
		} else {
			xb = qnx4fs_xalloc(ip, &bp, &xbnr);
			if (xb == NULL)
				goto fail;
			QNX4FS_I(ip)->i_xbnr0	= QNX4FS_I(ip)->i_xbnr_last
						= QNX4FS_I(ip)->i_xbnr
						= xbnr;
			xb->xb_xts[0].start	= cpu_to_le32(goal);
			xb->xb_xts[0].size	= xb->xb_nblocks
						= cpu_to_le32(1);
			xb->xb_xt0.start	= cpu_to_le32(QNX4FS_I(ip)->i_xt0.start);
			xb->xb_xt0.size		= cpu_to_le32(QNX4FS_I(ip)->i_xt0.size);
			++QNX4FS_I(ip)->i_nxt;
		}
		break;
	default:
		tmp = le32_to_cpu(xb->xb_nblocks) + 1;
		if (res) {
			xb->xb_nblocks = cpu_to_le32(tmp);
			tmp = le32_to_cpu(xb->xb_xts[ixt].size) + 1;
			xb->xb_xts[ixt].size = cpu_to_le32(tmp);
			if (QNX4FS_I(ip)->i_phys == le32_to_cpu(xb->xb_xts[ixt].start))
				++QNX4FS_I(ip)->i_offset_end;
			break;
		}
		if (++ixt == QNX4FS_MAX_XTS_PER_XBLK) {
			qnx4fs_xb_t *xb2;
			xb2 = qnx4fs_xalloc(ip, &bp2, &xbnr);
			if (xb2 == NULL) {
				brelse(bp);
				goto fail;
			}
			xb->xb_next = cpu_to_le32(xbnr);
			xb2->xb_prev = cpu_to_le32(QNX4FS_I(ip)->i_xbnr_last);
			QNX4FS_I(ip)->i_xbnr_last = xbnr;
			xb = xb2;
			ixt = 0;
			tmp = 1;
		}
		xb->xb_xts[ixt].start = cpu_to_le32(goal);
		xb->xb_xts[ixt].size = cpu_to_le32(1);
		xb->xb_nblocks = cpu_to_le32(tmp);
		if (bp2) {
			mark_buffer_dirty (bp2);
			brelse (bp2);
		} else {
			++xb->xb_nxt;
		}
		++QNX4FS_I(ip)->i_nxt;
	}

	++ip->i_blocks;
	QNX4FS_I(ip)->i_mmu_private += QNX4FS_BLOCK_SIZE;
	mark_inode_dirty(ip);
	if (bp) {
		mark_buffer_dirty(bp);
		brelse(bp);
	}
	return goal;
fail:
	return 0;
}
