/*
 *  linux/fs/qnx4fs/inode.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

static void
qnx4fs_count_iblocks (struct inode *ip)
{
	struct buffer_head *bp;
	qnx4fs_xb_t *xb;
	unsigned long iblocks, xbnr;

	iblocks = (ip->i_size + QNX4FS_BLOCK_SIZE-1) >> QNX4FS_BLOCK_SIZE_BITS;

	xbnr = QNX4FS_I(ip)->i_xbnr0;
	while (xbnr) {
		bp = bread(ip->i_dev, xbnr - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			continue;
		xb = (qnx4fs_xb_t *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse(bp);
			return;
		}
		++iblocks;
		QNX4FS_I(ip)->i_xbnr_last = xbnr;
		xbnr = le32_to_cpu(xb->xb_next);
		brelse(bp);
	}
	ip->i_blocks = iblocks;
}

void
qnx4fs_read_inode (struct inode *ip)
{
	struct buffer_head *bp;
	qnx4fs_dinode_t *qip;
	ino_t ino = ip->i_ino;

	bp = bread(ip->i_dev, ino2phys(ino), QNX4FS_BLOCK_SIZE);
	if (bp == NULL)
		goto fail;
	qip = (qnx4fs_dinode_t *)bp->b_data + ino2slot(ino);
	
	ip->i_mode	= le16_to_cpu(qip->i.mode);
	ip->i_uid	= (uid_t)le16_to_cpu(qip->i.uid);
	ip->i_gid	= (gid_t)le16_to_cpu(qip->i.gid);
        ip->i_nlink	= le16_to_cpu(qip->i.nlink);
	ip->i_size	= le32_to_cpu(qip->i.size);
	ip->i_mtime	= le32_to_cpu(qip->i.mtime);
        ip->i_atime	= le32_to_cpu(qip->i.atime);
	ip->i_ctime	= le32_to_cpu(qip->i.ctime);
	ip->i_blksize	= PAGE_CACHE_SIZE;

	QNX4FS_I(ip)->i_nxt		= le16_to_cpu(qip->i.nxt);
	QNX4FS_I(ip)->i_xbnr		= QNX4FS_I(ip)->i_xbnr0
					= le32_to_cpu(qip->i.xb);
	QNX4FS_I(ip)->i_phys		= QNX4FS_I(ip)->i_xt0.start
					= le32_to_cpu(qip->i.xt0.start);
	QNX4FS_I(ip)->i_offset_end	= QNX4FS_I(ip)->i_xt0.size
					= le32_to_cpu(qip->i.xt0.size);
	QNX4FS_I(ip)->i_offset		= 0;
	QNX4FS_I(ip)->i_xtnr		= -1;

	QNX4FS_I(ip)->i_linked		= qip->i.status & QNX4FS_FILE_INODE;
	QNX4FS_I(ip)->i_mmu_private	= ip->i_size;

	brelse(bp);
	if (S_ISREG(ip->i_mode)) {
		ip->i_op = &qnx4fs_file_inode_ops;
		ip->i_fop = &qnx4fs_file_ops;
		ip->i_mapping->a_ops = &qnx4fs_address_space_ops;
	} else if (S_ISDIR(ip->i_mode)) {
		ip->i_op = &qnx4fs_dir_inode_ops;
		ip->i_fop = &qnx4fs_dir_ops;
		ip->i_mapping->a_ops = &qnx4fs_address_space_ops;
	} else if (S_ISLNK(ip->i_mode)) {
		ip->i_op = &page_symlink_inode_operations;
		ip->i_mapping->a_ops = &qnx4fs_address_space_ops;
	} else {
		goto fail;
	}
	qnx4fs_count_iblocks(ip);
	return;
fail:
	make_bad_inode(ip);
}

void
qnx4fs_write_inode (struct inode *ip, int wait)
{
	struct buffer_head *bp;
	qnx4fs_dinode_t *qip;
	ino_t ino = ip->i_ino;

	lock_kernel();
	bp = bread(ip->i_dev, ino2phys(ino), QNX4FS_BLOCK_SIZE);
	if (bp == NULL)
		goto out;
	qip = (qnx4fs_dinode_t *)bp->b_data + ino2slot(ino);

	qip->i.size	 = cpu_to_le32(ip->i_size);
	qip->i.mode	 = cpu_to_le16(ip->i_mode);
	qip->i.uid	 = cpu_to_le16(ip->i_uid);
	qip->i.gid	 = cpu_to_le16(ip->i_gid);
	qip->i.nlink	 = cpu_to_le16(ip->i_nlink);
	qip->i.atime	 = cpu_to_le32(ip->i_atime);
	qip->i.ctime	 = cpu_to_le32(ip->i_ctime);
	qip->i.mtime	 = cpu_to_le32(ip->i_mtime);
	qip->i.xt0.start = cpu_to_le32(QNX4FS_I(ip)->i_xt0.start);
	qip->i.xt0.size	 = cpu_to_le32(QNX4FS_I(ip)->i_xt0.size);
	qip->i.xb	 = cpu_to_le32(QNX4FS_I(ip)->i_xbnr0);
	qip->i.nxt	 = cpu_to_le16(QNX4FS_I(ip)->i_nxt);
	if (QNX4FS_I(ip)->i_linked)
		qip->i.status |= QNX4FS_FILE_INODE;

	mark_buffer_dirty(bp);
	brelse(bp);
out:
	unlock_kernel();
}

void
qnx4fs_delete_inode (struct inode *ip)
{
	lock_kernel();
	ip->i_size = 0;
	qnx4fs_truncate(ip);
	if (QNX4FS_I(ip)->i_linked)
		qnx4fs_difree (ip->i_dev, ip->i_ino);
	unlock_kernel();
	clear_inode(ip);
}
