#ifndef __QNX4FS_FS_H
#define __QNX4FS_FS_H

#define QNX4FS_ROOT_INO			8
#define QNX4FS_INODES_INO		9
#define QNX4FS_BOOT_INO			10
#define QNX4FS_ALTBOOT_INO		11

#define QNX4FS_LINK_MAX			65535

#define QNX4FS_SUPER_MAGIC		0x34584e51 /* little endian "QNX4" */

#define QNX4FS_BLOCK_SIZE		512
#define QNX4FS_BLOCK_SIZE_BITS		9
#define QNX4FS_DIRENT_SIZE		64
#define QNX4FS_DIRENT_SIZE_BITS		6
#define QNX4FS_INODES_PER_BLOCK		8
#define QNX4FS_INODES_PER_BLOCK_BITS	3
#define QNX4FS_DIRENTS_PER_BLOCK	QNX4FS_INODES_PER_BLOCK
#define QNX4FS_DIRENTS_PER_BLOCK_BITS	QNX4FS_INODES_PER_BLOCK_BITS

#define QNX4FS_BOOT_BLOCK		1L
#define QNX4FS_ROOT_BLOCK		2L
#define QNX4FS_BITMAP_BLOCK		3L

#define QNX4FS_MAX_XTS_PER_XBLK		60
#define QNX4FS_LONG_NAME_MAX		48
#define QNX4FS_SHORT_NAME_MAX		16

#define QNX4FS_FILE_USED		0x01
#define QNX4FS_FILE_MODIFIED		0x02
#define QNX4FS_FILE_BUSY		0x04
#define QNX4FS_FILE_LINK		0x08
#define QNX4FS_FILE_INODE		0x10
#define QNX4FS_FSYS_CLEAN		0x20
#define QNX4FS_FILE_GROWN		0x40

#define QNX4FS_DIR_EXPAND_DELTA		4
#define QNX4FS_INODES_EXPAND_DELTA	16

typedef struct qnx4fs_xb {
	u32		xb_next;
	u32		xb_prev;
	u8		xb_nxt;
	char		xb_spare[3];
	s32		xb_nblocks;
	qnx4fs_xt_t	xb_xts[QNX4FS_MAX_XTS_PER_XBLK];
	char		xb_signature[8];
	qnx4fs_xt_t	xb_xt0;
} qnx4fs_xb_t;

struct qnx4fs_dinode {
	char		name[QNX4FS_SHORT_NAME_MAX];
	u32		size;
	qnx4fs_xt_t	xt0;
	u32		xb;
	u32		ftime;
	u32		mtime;
	u32		atime;
	u32		ctime;
	u16		nxt;
	u16		mode;
	u16		uid;
	u16		gid;
	u16		nlink;
	char		zero[4];
	u8		type;
	u8		status;
};

struct qnx4fs_dlink {
	char	name[QNX4FS_LONG_NAME_MAX];
	u32	block;
	u8	index;
	char	spare[10];
	u8	status;
};

typedef union qnx4fs_dirent  {
	struct qnx4fs_dinode i;
	struct qnx4fs_dlink l;
} qnx4fs_dirent_t;

typedef qnx4fs_dirent_t qnx4fs_dinode_t;

/* we don't need this */
typedef struct qnx4fs_super_block {
	qnx4fs_dinode_t root;
	qnx4fs_dinode_t inodes;
	qnx4fs_dinode_t boot;
	qnx4fs_dinode_t altboot;
} qnx4fs_super_block_t;

#ifdef  QNX4DBG
#define QNXDBG(x)  do{printk("qnx4fs:" __FUNCTION__); printk x ;} while(0)
#else
#define QNXDBG(x)   (void)0
#endif

#ifdef __KERNEL__

#define QNX4FS_I(ip)	(&(ip)->u.qnx4fs_i)
#define QNX4FS_SB(sbp)	(&(sbp)->u.qnx4fs_sb)

/* bitmap.c */
extern void qnx4fs_bfree (struct super_block *, unsigned long, unsigned long);
extern int qnx4fs_balloc (struct super_block *, unsigned long *);
extern unsigned long qnx4fs_count_free_blocks (struct super_block *);

/* common.c */
extern inline void qnx4fs_inc_nlink (struct inode *);
extern inline void qnx4fs_dec_nlink (struct inode *);
extern inline ino_t direct2ino (unsigned long, int);
extern inline ino_t link2ino (qnx4fs_dirent_t *);
extern inline unsigned long ino2phys (ino_t);
extern inline int ino2slot (ino_t);
extern inline long qnx4fs_nblocks (struct inode *);
extern struct buffer_head *qnx4fs_get_slot (struct inode *, ino_t *,
					    qnx4fs_dirent_t **, int);
extern void qnx4fs_mark_slot (qnx4fs_dirent_t *, struct buffer_head *, int);

/* dir.c */
extern struct file_operations qnx4fs_dir_ops;

/* file.c */
extern struct file_operations qnx4fs_file_ops;
extern struct inode_operations qnx4fs_file_inode_ops;
extern struct address_space_operations qnx4fs_address_space_ops;

extern int qnx4fs_get_block (struct inode *, long,
			     struct buffer_head *, int);
extern unsigned long qnx4fs_map_block (struct inode *, long);

/* ialloc.c */
extern int qnx4fs_dialloc (struct super_block *, ino_t *);
extern void qnx4fs_difree (kdev_t, ino_t);
extern struct buffer_head *qnx4fs_inew (struct inode *, struct dentry *,
					struct inode **, qnx4fs_dirent_t **);
extern unsigned long qnx4fs_iexpand (struct inode *);
extern struct buffer_head *qnx4fs_iexpand_special (struct inode *, int);

/* inode.c */
extern struct kmem_cache_s *qnx4fs_inode_cachep;

extern void qnx4fs_read_inode (struct inode *);
extern void qnx4fs_write_inode (struct inode *, int);
extern void qnx4fs_put_inode (struct inode *);
extern void qnx4fs_delete_inode (struct inode *);

/* namei.c */
extern struct inode_operations qnx4fs_dir_inode_ops;

/* super.c */
extern struct super_operations qnx4fs_super_ops;

/* truncate.c */
extern void qnx4fs_truncate (struct inode *);

#endif /* __KERNEL__ */

#endif /* __QNX4FS_FS_H */
