/*
 *  linux/fs/qnx4fs/namei.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  01.10.2002. - Dmitry Yakovlev - some fixes.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

static inline void
qnx4fs_link_dirent (qnx4fs_dirent_t *qdp, ino_t ino)
{
	qdp->l.block = cpu_to_le32(ino2phys(ino) + 1);
	qdp->l.index = ino2slot(ino);
	qdp->l.status = QNX4FS_FILE_LINK;
}

static qnx4fs_dirent_t *
qnx4fs_find_dirent (struct inode *dip, struct dentry *dp,
		    struct buffer_head **bpp, ino_t *inop)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long block, nblocks;
	const char *name = dp->d_name.name;
	int slot, namelen = dp->d_name.len;
	ino_t ino;

	nblocks = qnx4fs_nblocks(dip);
	for (block = 0; block < nblocks; block++) {
		phys = qnx4fs_map_block(dip, block);
		if (!phys)
			continue;
		bp = bread(dip->i_dev, phys - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			continue;
		qdp = (qnx4fs_dirent_t *)bp->b_data;
		for (slot = 0; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (qdp->l.status & QNX4FS_FILE_LINK) {
				if (strncmp(qdp->l.name, name,
					    QNX4FS_LONG_NAME_MAX) == 0) {
					ino = link2ino(qdp);
					goto found;
				}
			} else if ((qdp->i.status & QNX4FS_FILE_USED)
				   && namelen <= QNX4FS_SHORT_NAME_MAX
				   && strncmp(qdp->i.name, name,
					       QNX4FS_SHORT_NAME_MAX) == 0) {
				ino = direct2ino(phys, slot);
				goto found;
			}
		}
		brelse(bp);
	}

	return NULL;
found:
	if (ino == QNX4FS_INODES_INO
	    || ino == QNX4FS_SB(dip->i_sb)->s_bitmap_ino) {
	    	brelse(bp);
		return NULL;
	}
	*inop = ino;
	*bpp = bp;
	return qdp;
}

static int
qnx4fs_is_dir_empty (struct inode *dip)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	long block, nblocks;
	unsigned long phys;
	int slot = 2;

	nblocks = qnx4fs_nblocks(dip);
	for (block = 0; block < nblocks; block++) {
		phys = qnx4fs_map_block(dip, block);
		if (!phys)
			return 0;
		bp = bread(dip->i_dev, phys - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL)
			return 0;
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (qdp->i.status) {
				brelse(bp);
				return 0;
			}
		}
		slot = 0;
		brelse(bp);
	}

	return 1;
}

static int
qnx4fs_make_empty_dir (struct inode *dip, struct inode *ip)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;

	bp = qnx4fs_iexpand_special(ip, 4);
	if (bp == NULL)
		return -ENOSPC;
	qdp = (qnx4fs_dirent_t *)bp->b_data;

	memcpy(qdp->l.name, ".\0I\3QNX\0\4", 9);
	/* This makes chkfsys happy... */
	qdp->l.name[8] = QNX4FS_I(ip)->i_xt0.size;
	qnx4fs_link_dirent(qdp++, ip->i_ino);
	memcpy(qdp->l.name, "..\0", 3);

	qnx4fs_link_dirent(qdp, dip->i_ino);
	qnx4fs_inc_nlink(ip);
	mark_buffer_dirty(bp);
	brelse(bp);

	return 0;
}

static int
qnx4fs_fill_dirent (qnx4fs_dirent_t *qdp, struct buffer_head *bp,
		    struct dentry *dp, struct inode *ip)
{
	struct inode *dip = dp->d_parent->d_inode;

	memcpy(qdp->i.name, dp->d_name.name, dp->d_name.len + 1);
	if (QNX4FS_I(ip)->i_linked)
		qnx4fs_link_dirent(qdp, ip->i_ino);
	else
		qdp->i.status = QNX4FS_FILE_USED;

	dip->i_mtime = dip->i_ctime = CURRENT_TIME;
	mark_inode_dirty (dip);
	mark_buffer_dirty(bp);
	brelse(bp);

	return 0;
}

static inline int
qnx4fs_link_inode (struct inode *ip, ino_t new_ino)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;

	ino = ip->i_ino;
	bp = bread(ip->i_dev, ino2phys(ino), QNX4FS_BLOCK_SIZE);
	if (bp == NULL)
		return -EIO;
	qdp = (qnx4fs_dirent_t *)bp->b_data + ino2slot(ino);
	qnx4fs_link_dirent(qdp, new_ino);
	mark_buffer_dirty(bp);
	brelse(bp);
	ip->i_ino = new_ino;
	QNX4FS_I(ip)->i_linked = 1;

	return 0;
}

static int
qnx4fs_instantiate_dirent (qnx4fs_dirent_t *qdp, struct buffer_head *bp,
			   struct dentry *dp, struct inode *ip)
{
	int err;

	err = qnx4fs_fill_dirent(qdp, bp, dp, ip);
	if (err) {
		ip->i_nlink = 0;
		mark_inode_dirty(ip);
		iput(ip);
	} else {
		mark_inode_dirty(ip);
		d_instantiate(dp, ip);
	}
	return err;
}

static struct dentry *
qnx4fs_lookup (struct inode *dip, struct dentry *dp)
{
	struct inode *ip;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;

	if (dp->d_name.len > QNX4FS_LONG_NAME_MAX)
		return ERR_PTR(-ENAMETOOLONG);

	qdp = qnx4fs_find_dirent(dip, dp, &bp, &ino);
	if (qdp == NULL) {
		ip = NULL;
	} else {
		brelse(bp);
		ip = iget(dip->i_sb, ino);
		if (ip == NULL)
			return ERR_PTR(-EACCES);
	}

	d_add(dp, ip);
	return NULL;
}

static int
qnx4fs_create (struct inode *dip, struct dentry *dp, int mode)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	struct inode *ip;

	bp = qnx4fs_inew(dip, dp, &ip, &qdp);
	if (IS_ERR(bp))
		return PTR_ERR(bp);

	ip->i_mode		= mode;
	ip->i_op		= &qnx4fs_file_inode_ops;
	ip->i_fop		= &qnx4fs_file_ops;
	ip->i_mapping->a_ops	= &qnx4fs_address_space_ops;

	insert_inode_hash(ip);
	return qnx4fs_instantiate_dirent(qdp, bp, dp, ip);
}

static int
qnx4fs_link (struct dentry *old_dp, struct inode *dip, struct dentry *dp)
{
	struct inode *ip = old_dp->d_inode;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;
	int err;

	if (S_ISDIR(ip->i_mode))
		return -EPERM;
	if (ip->i_nlink >= QNX4FS_LINK_MAX)
		return -EMLINK;

	if (!QNX4FS_I(ip)->i_linked) {
		err = qnx4fs_dialloc(ip->i_sb, &ino);
		if (err)
			goto out;
		err = qnx4fs_link_inode(ip, ino);
		if (err) {
			qnx4fs_difree(ip->i_dev, ino);
			goto out;
		}
	}
	bp = qnx4fs_get_slot(dip, &ino, &qdp, QNX4FS_DIR_EXPAND_DELTA);
	if (bp == NULL) {
		err = -ENOSPC;
		goto out;
	}
	err = qnx4fs_fill_dirent(qdp, bp, dp, ip);
	if (!err) {
		ip->i_ctime = CURRENT_TIME;
		qnx4fs_inc_nlink(ip);
		atomic_inc(&ip->i_count);
		d_instantiate(dp, ip);
	}
out:
	return err;
}

static int
qnx4fs_unlink (struct inode *dip, struct dentry *dp)
{
	struct inode *ip = dp->d_inode;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t unused;

	qdp = qnx4fs_find_dirent(dip, dp, &bp, &unused);
	if (qdp == NULL)
		return -ENOENT;
	qnx4fs_mark_slot(qdp, bp, 0);
	ip->i_ctime = dip->i_ctime;
	qnx4fs_dec_nlink(ip);

	return 0;
}

static int
qnx4fs_symlink (struct inode *dip, struct dentry *dp, const char *symname)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	struct inode *ip;
	int err, len = strlen(symname);

	if (len > QNX4FS_BLOCK_SIZE)
		return -ENAMETOOLONG;

	bp = qnx4fs_inew(dip, dp, &ip, &qdp);
	if (IS_ERR(bp))
		return PTR_ERR(bp);

	ip->i_mode		= S_IFLNK | S_IRWXUGO;
	ip->i_op		= &page_symlink_inode_operations;
	ip->i_mapping->a_ops	= &qnx4fs_address_space_ops;

	insert_inode_hash(ip);
	err = block_symlink(ip, symname, len + 1);
	if (err) {
		brelse(bp);
		qnx4fs_dec_nlink(ip);
		iput(ip);
		return err;
	}
	return qnx4fs_instantiate_dirent (qdp, bp, dp, ip);
}

static int
qnx4fs_mkdir (struct inode *dip, struct dentry *dp, int mode)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	struct inode *ip;
	int err;

	if (dip->i_nlink >= QNX4FS_LINK_MAX)
		return -EMLINK;

	qnx4fs_inc_nlink(dip);
	bp = qnx4fs_inew(dip, dp, &ip, &qdp);
	if (IS_ERR(bp)) {
		err = PTR_ERR(bp);
		goto fail;
	}

	ip->i_mode		= S_IFDIR | mode;
	ip->i_op		= &qnx4fs_dir_inode_ops;
	ip->i_fop		= &qnx4fs_dir_ops;
	ip->i_mapping->a_ops	= &qnx4fs_address_space_ops;

	insert_inode_hash(ip);
	err = qnx4fs_make_empty_dir(dip, ip);
	if (err) {
		brelse(bp);
		qnx4fs_dec_nlink(ip);
		iput(ip);
		goto fail;
	}

	err = qnx4fs_instantiate_dirent(qdp, bp, dp, ip);
	if (!err)
		return 0;
fail:
	qnx4fs_dec_nlink(dip);
	return err;
}

static int
qnx4fs_rmdir (struct inode *dip, struct dentry *dp)
{
	struct inode *ip = dp->d_inode;
	int err = -ENOTEMPTY;

	if (qnx4fs_is_dir_empty(ip)) {
		err = qnx4fs_unlink(dip, dp);
		if (!err) {
			qnx4fs_dec_nlink(dip);
			qnx4fs_dec_nlink(ip);
		}
	}

	return err;
}

static int
qnx4fs_rename_helper (struct inode *ip, struct dentry *dp, ino_t *inop)
{
	int err;

	if (QNX4FS_I(ip)->i_linked) {
		inop = &ip->i_ino;
	} else if (dp->d_name.len > QNX4FS_SHORT_NAME_MAX) {
		err = qnx4fs_dialloc(ip->i_sb, inop);
		if (err) {
			return err;
		}
		QNX4FS_I(ip)->i_linked = 1;
	}
	ip->i_ino = *inop;

	return 0;
}

static int
qnx4fs_rename (struct inode *old_dip, struct dentry *old_dp,
	       struct inode *new_dip, struct dentry *new_dp)
{
	struct inode *old_ip = old_dp->d_inode;
	struct inode *new_ip = new_dp->d_inode;
	struct buffer_head *old_bp, *new_bp, *dir_bp = NULL;
	qnx4fs_dirent_t *old_qdp, *new_qdp, *dir_qdp = NULL;
	ino_t ino;
	int err;

	old_qdp = qnx4fs_find_dirent (old_dip, old_dp, &old_bp, &ino);
	if (old_qdp == NULL)
		return -ENOENT;

	if (S_ISDIR(old_ip->i_mode)) {
		dir_bp = bread(old_ip->i_dev, 0, QNX4FS_BLOCK_SIZE);
		if (dir_bp == NULL) {
			err = -EIO;
			goto out_old_bp;
		}
		dir_qdp = (qnx4fs_dirent_t *)dir_bp->b_data;
	}

	if (new_ip == NULL) {
		if (dir_qdp && new_dip->i_nlink >= QNX4FS_LINK_MAX) {
			err = -EMLINK;
			goto out_old_bp;
		}
		qnx4fs_inc_nlink(old_ip);
		new_bp = qnx4fs_get_slot(new_dip, &ino, &new_qdp,
					 QNX4FS_DIR_EXPAND_DELTA);
		if (IS_ERR(new_bp)) {
			qnx4fs_dec_nlink(old_ip);
			err = PTR_ERR(new_bp);
			goto out_old_bp;
		}
		err = qnx4fs_rename_helper(old_ip, new_dp, &ino);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			brelse(new_bp);
			goto out_old_bp;
		}
		err = qnx4fs_fill_dirent(new_qdp, new_bp, new_dp, old_ip);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			goto out_old_bp;
		}
		if (dir_qdp)
			qnx4fs_inc_nlink(new_dip);
	} else {
		if (dir_qdp && !qnx4fs_is_dir_empty(new_ip)) {
			err = -ENOTEMPTY;
			goto out_old_bp;
		}
		new_qdp = qnx4fs_find_dirent(new_dip, new_dp, &new_bp, &ino);
		if (new_qdp == NULL) {
			err = -ENOENT;
			goto out_old_bp;
		}
		qnx4fs_inc_nlink(old_ip);
		err = qnx4fs_rename_helper(old_ip, new_dp, &ino);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			brelse(new_bp);
			goto out_old_bp;
		}
		err = qnx4fs_fill_dirent(new_qdp, new_bp, new_dp, old_ip);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			goto out_old_bp;
		}
		new_ip->i_ctime = CURRENT_TIME;
		if (dir_qdp)
			new_ip->i_nlink--;
		qnx4fs_dec_nlink(new_ip);
	}
	qnx4fs_mark_slot(old_qdp, old_bp, 0);
	qnx4fs_dec_nlink(old_ip);

	if (dir_qdp) {
		qnx4fs_dec_nlink(old_dip);
		qnx4fs_link_dirent(dir_qdp++, old_ip->i_ino);
		qnx4fs_link_dirent(dir_qdp, new_dip->i_ino);
		mark_buffer_dirty(dir_bp);
		brelse(dir_bp);
	}

	return 0;
out_old_bp:
	brelse(old_bp);
	if (dir_qdp)
		brelse(dir_bp);

	return err;
}

struct inode_operations
qnx4fs_dir_inode_ops = {
	create:		qnx4fs_create,
	lookup:		qnx4fs_lookup,
	link:		qnx4fs_link,
	unlink:		qnx4fs_unlink,
	symlink:	qnx4fs_symlink,
	mkdir:		qnx4fs_mkdir,
	rmdir:		qnx4fs_rmdir,
	rename:		qnx4fs_rename
};
