/*
 *  linux/fs/qnx4fs/super.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  09.10.2002 Dmitry Yakovlev - some fixes. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

static struct super_block *
qnx4fs_read_super (struct super_block *sbp, void *dp, int silent)
{
	kdev_t dev = sbp->s_dev;
	struct buffer_head *bp;
	struct inode *ip;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;

	set_blocksize(dev, QNX4FS_BLOCK_SIZE);
	sbp->s_blocksize = QNX4FS_BLOCK_SIZE;
	sbp->s_blocksize_bits = QNX4FS_BLOCK_SIZE_BITS;

	bp = bread (dev, QNX4FS_ROOT_BLOCK - 1, QNX4FS_BLOCK_SIZE);
	if (bp == NULL) {
		printk(KERN_WARNING "qnx4fs: unable to read root block\n");
		goto out;
	}
	qdp = (qnx4fs_dirent_t *)bp->b_data;

	if (memcmp(qdp->i.name, "/", 2)) {
		if (!silent)
			printk(KERN_WARNING "qnx4fs: no root dir\n");
		goto outbp;
	}

	phys = le32_to_cpu(qdp->i.xt0.start);
	QNX4FS_SB(sbp)->s_bitmap_ino = direct2ino(phys, 2);
	QNX4FS_SB(sbp)->s_bitmap_blocks = phys - QNX4FS_BITMAP_BLOCK;
	brelse (bp);

	bp = bread(dev, phys - 1, QNX4FS_BLOCK_SIZE);
	if (bp == NULL) {
		printk(KERN_WARNING "qnx4fs: unable to read root dir block\n");
		goto out;
	}
	qdp = (qnx4fs_dirent_t *)bp->b_data + 2;
	if (memcmp(qdp->i.name, ".bitmap\0", 8)) {
		if (!silent)
			printk(KERN_WARNING "qnx4fs: wrong '/.bitmap' entry\n");
		goto outbp;
	}
	QNX4FS_SB(sbp)->s_blocks = le32_to_cpu(qdp->i.size) << 3;
	brelse(bp);

	sbp->s_op = &qnx4fs_super_ops;
	sbp->s_magic = QNX4FS_SUPER_MAGIC;
	/* sbp->s_flags = MS_RDONLY; */

	sbp->s_root = d_alloc_root(iget(sbp, QNX4FS_ROOT_INO));
	if (sbp->s_root == NULL) {
		printk(KERN_WARNING "qnx4fs: unable to get root dentry\n");
		goto out;
	}

	ip = iget(sbp, QNX4FS_INODES_INO);
	if (ip == NULL) {
		printk(KERN_WARNING "qnx4fs: unable to get '/.inodes' inode\n");
		iput(sbp->s_root->d_inode);
		goto out;
	}
	QNX4FS_SB(sbp)->s_inodes = ip;

	if (!(sbp->s_flags & MS_RDONLY))
		sbp->s_dirt = 1;

	return sbp;
outbp:
	brelse(bp);
out:
	return NULL;
}

static void
qnx4fs_put_super (struct super_block *sbp)
{
	iput(QNX4FS_SB(sbp)->s_inodes);
}

static void
qnx4fs_write_super (struct super_block *sbp)
{
	sbp->s_dirt = 0;
}

static int
qnx4fs_remount_fs (struct super_block *sbp, int *flags, char *dp)
{
	return 0;
}

static int
qnx4fs_statfs (struct super_block *sbp, struct statfs *stat)
{
	stat->f_type	= sbp->s_magic;
	stat->f_bsize	= sbp->s_blocksize;
	stat->f_blocks	= QNX4FS_SB(sbp)->s_blocks;
	stat->f_bfree	= stat->f_bavail
			= qnx4fs_count_free_blocks(sbp);
	stat->f_files	= 0;
	stat->f_ffree	= stat->f_bfree;
	stat->f_namelen	= QNX4FS_LONG_NAME_MAX;

	return 0;
}

struct super_operations
qnx4fs_super_ops = {
	read_inode:	qnx4fs_read_inode,
	write_inode:	qnx4fs_write_inode,
	delete_inode:	qnx4fs_delete_inode,
	put_super:	qnx4fs_put_super,
	write_super:	qnx4fs_write_super,
	statfs:		qnx4fs_statfs,
	remount_fs:	qnx4fs_remount_fs
};

static DECLARE_FSTYPE_DEV(qnx4fs_fs_type, "qnx4fs", qnx4fs_read_super);

static int __init
qnx4fs_init (void)
{
	return register_filesystem(&qnx4fs_fs_type);
}

static void __exit
qnx4fs_cleanup (void)
{
	unregister_filesystem(&qnx4fs_fs_type);
}

MODULE_AUTHOR ("Serguei Tzukanov <tzukanov@narod.ru>");
MODULE_DESCRIPTION ("QNX4 filesystem driver for Linux");
MODULE_LICENSE ("GPL");
EXPORT_NO_SYMBOLS;

module_init (qnx4fs_init);
module_exit (qnx4fs_cleanup);
