/*
 *  linux/fs/qnx4fs/truncate.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  05.11.2002  Dmitry Yakovlev - qnx4fs_truncate() fixes 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/pagemap.h>
#include <linux/fs.h>
#include <linux/qnx4fs_fs.h>

static inline void
free_extent_le (struct super_block *sbp, qnx4fs_xt_t *xt)
{
	qnx4fs_bfree(sbp, le32_to_cpu(xt->start), le32_to_cpu(xt->size));
}

static inline void
shorten_extent_by (struct super_block *sbp, qnx4fs_xt_t *xt, int size)
{
	qnx4fs_bfree(sbp, le32_to_cpu(xt->start) 
		+ le32_to_cpu(xt->size) - size, size);
	xt->size = cpu_to_le32(le32_to_cpu(xt->size) - size);
}

void
qnx4fs_truncate (struct inode *ip)
{
	struct buffer_head *bp;
	struct super_block *sbp = ip->i_sb;
	qnx4fs_xb_t *xb;
	qnx4fs_xt_t *xt;
	unsigned long xbnr, xbnr_tmp;
	unsigned long isize, offset;
	unsigned long nblocks;
	int nxt, nxt_rest;

	/* this is stolen from fs/affs/file.c:affs_truncate */
	if (ip->i_size > QNX4FS_I(ip)->i_mmu_private) {
		struct address_space *mapping = ip->i_mapping;
		struct page *page;
		unsigned long size = ip->i_size - 1;

		page = grab_cache_page(mapping, size >> PAGE_CACHE_SHIFT);
		if (page == NULL)
			return;
		size = (size & (PAGE_CACHE_SIZE - 1)) + 1;
		if (!mapping->a_ops->prepare_write(NULL, page, size, size))
			mapping->a_ops->commit_write(NULL, page, size, size);
		UnlockPage(page);
		page_cache_release(page);
		mark_inode_dirty(ip);
		return;
	}

	isize = (ip->i_size + QNX4FS_BLOCK_SIZE-1) >> QNX4FS_BLOCK_SIZE_BITS;
	nxt_rest = QNX4FS_I(ip)->i_nxt;
	offset = QNX4FS_I(ip)->i_xt0.size;
	xbnr = QNX4FS_I(ip)->i_xbnr0;
	if (isize <= offset) {
		qnx4fs_bfree(sbp, QNX4FS_I(ip)->i_xt0.start + isize,
			offset - isize);
		QNX4FS_I(ip)->i_xt0.size = isize;
		QNX4FS_I(ip)->i_xbnr0 = 0;
		QNX4FS_I(ip)->i_xbnr_last = 0;
		if (nxt_rest && !isize){
			QNX4FS_I(ip)->i_xt0.start = 0;
			nxt_rest--;
		}
	}
	QNX4FS_I(ip)->i_offset		= 0;
	QNX4FS_I(ip)->i_offset_end	= QNX4FS_I(ip)->i_xt0.size;
	QNX4FS_I(ip)->i_phys		= QNX4FS_I(ip)->i_xt0.start;
	QNX4FS_I(ip)->i_xtnr		= -1;
	QNX4FS_I(ip)->i_xbnr		= QNX4FS_I(ip)->i_xbnr0;
	for (; xbnr; brelse (bp)) {
		bp = bread (ip->i_dev, xbnr - 1, QNX4FS_BLOCK_SIZE);
		if (bp == NULL) {
			printk (KERN_WARNING
				"qnx4fs: unable to finish truncate ().\n");
			goto out;
		}
		xb = (qnx4fs_xb_t *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse(bp);
			goto out;
		}
		if (isize <= offset) {
			nxt = xb->xb_nxt;
			nxt_rest -= nxt;
			xt = xb->xb_xts;
			for (; nxt > 0; nxt--) {
				free_extent_le(sbp, xt);
				xt++;
			}
			xbnr_tmp = xbnr;
			xbnr = le32_to_cpu(xb->xb_next);
			qnx4fs_bfree(sbp, xbnr_tmp, 1);
			continue;
		}
		nblocks = le32_to_cpu(xb->xb_nblocks);
		offset += nblocks;
		if (isize > offset) {
			xbnr = le32_to_cpu(xb->xb_next);
			continue;
		}
		nxt = xb->xb_nxt;
		xt = &xb->xb_xts[nxt - 1];
		for (; nxt > 0; nxt--) {
			int size = le32_to_cpu(xt->size);

			if (isize > (offset - size)) {
				size = offset - isize;
				shorten_extent_by(sbp, xt, size);
				nblocks -= size;
				nxt--;
				break;
			}
			offset -= size;
			nblocks -= size;
			free_extent_le(sbp, xt);
			xt->start = 0;
			xt->size  = 0;
			nxt_rest--;
			xt--;
		}
		offset = isize;  // ???
		xb->xb_nxt = nxt + 1;
		xb->xb_nblocks = cpu_to_le32(nblocks);
		QNX4FS_I(ip)->i_xbnr_last = xbnr;
		xbnr = le32_to_cpu(xb->xb_next);
		xb->xb_next = 0;
		mark_buffer_dirty(bp);
	}
out:
	QNX4FS_I(ip)->i_nxt = nxt_rest;
	QNX4FS_I(ip)->i_mmu_private = ip->i_size;
	ip->i_mtime = CURRENT_TIME;
	mark_inode_dirty(ip);
}
